@@ -5,6 +5,7 @@ import math
 from typing import List, Callable
 from coopstructs.toggles import EnumToggleable
 from coopstructs.geometry import Rectangle
+import numpy as np
 
 class Orientation(Enum):
     UP_LEFT = 3
@@ -338,8 +339,8 @@ class CatmullRom(Curve):
     def __init__(self, id:str, control_points: List[Vector2]):
         Curve.__init__(self, id, control_points[0])
 
-        if len(control_points) < 3:
-            raise ValueError(f"Invalid input for control points. Must have at least 3 values but list of lenght {len(control_points)} was provided: [{control_points}]")
+        if len(control_points) < 4:
+            raise ValueError(f"Invalid input for control points. Must have at least 4 values but list of lenght {len(control_points)} was provided: [{control_points}]")
         self.control_points = control_points
 
     def line_representation(self):
@@ -349,7 +350,7 @@ class CatmullRom(Curve):
         return self.control_points[-1]
 
     def compute_curve_lines(self):
-        b_points = self.compute_bezier_points([(x.x, x.y) for x in self.control_points])
+        b_points = self.compute_catmull_points([(x.x, x.y) for x in self.control_points])
         ret = []
         for ii in range(0, len(b_points)):
             if ii == 0:
@@ -360,73 +361,63 @@ class CatmullRom(Curve):
                 ret.append(Line(self.id + f"_{ii}", start, end))
         return ret
 
-    def compute_bezier_points(self, vertices, numPoints=None):
+    def compute_catmull_points(self, vertices: List[Vector2], numPoints=None):
         if numPoints is None:
             numPoints = 30
-        if numPoints < 2 or len(vertices) != 3:
+        if numPoints < 2 or len(vertices) < 4:
             return []
 
-        result = []
+        ##### MODIFIED FROM: https://en.wikipedia.org/wiki/Centripetal_Catmull%E2%80%93Rom_spline
+
+        # The curve c will contain an array of (x, y) points.
+        c = []
+        for ii in range(len(vertices) - 3):
+            # Convert the points to numpy so that we can do array multiplication
+            P0, P1, P2, P3 = map(np.array, [(self.control_points[ii].x, self.control_points[ii].y),
+                                            (self.control_points[ii + 1].x, self.control_points[ii + 1].y),
+                                            (self.control_points[ii + 2].x, self.control_points[ii + 2].y),
+                                            (self.control_points[ii + 3].x, self.control_points[ii + 3].y)])
+
+            # Parametric constant: 0.5 for the centripetal spline, 0.0 for the uniform spline, 1.0 for the chordal spline.
+            alpha = 0.5
+            # Premultiplied power constant for the following tj() function.
+            alpha = alpha / 2
+
+            def tj(ti, Pi, Pj):
+                xi, yi = Pi
+                xj, yj = Pj
+                return ((xj - xi) ** 2 + (yj - yi) ** 2) ** alpha + ti
+
+            # Calculate t0 to t4
+            t0 = 0
+            t1 = tj(t0, P0, P1)
+            t2 = tj(t1, P1, P2)
+            t3 = tj(t2, P2, P3)
+
+            # Only calculate points between P1 and P2
+            t = np.linspace(t1, t2, numPoints)
+
+            # Reshape so that we can multiply by the points P0 to P3
+            # and get a point for each value of t.
+            t = t.reshape(len(t), 1)
+            # print(t)
+            A1 = (t1 - t) / (t1 - t0) * P0 + (t - t0) / (t1 - t0) * P1
+            A2 = (t2 - t) / (t2 - t1) * P1 + (t - t1) / (t2 - t1) * P2
+            A3 = (t3 - t) / (t3 - t2) * P2 + (t - t2) / (t3 - t2) * P3
+            # print(A1)
+            # print(A2)
+            # print(A3)
+            B1 = (t2 - t) / (t2 - t0) * A1 + (t - t0) / (t2 - t0) * A2
+            B2 = (t3 - t) / (t3 - t1) * A2 + (t - t1) / (t3 - t1) * A3
+
+            C = (t2 - t) / (t2 - t1) * B1 + (t - t1) / (t2 - t1) * B2
+
+            c.extend(C)
 
 
+        return [Vector2(point[0], point[1]) for point in c]
 
 
-        b0x = vertices[0][0]
-        b0y = vertices[0][1]
-        b1x = vertices[1][0]
-        b1y = vertices[1][1]
-        b2x = vertices[2][0]
-        b2y = vertices[2][1]
-        b3x = vertices[3][0]
-        b3y = vertices[3][1]
-
-        # Compute polynomial coefficients from Bezier points
-        ax = -b0x + 3 * b1x + -3 * b2x + b3x
-        ay = -b0y + 3 * b1y + -3 * b2y + b3y
-
-        bx = 3 * b0x + -6 * b1x + 3 * b2x
-        by = 3 * b0y + -6 * b1y + 3 * b2y
-
-        cx = -3 * b0x + 3 * b1x
-        cy = -3 * b0y + 3 * b1y
-
-        dx = b0x
-        dy = b0y
-
-        # Set up the number of steps and step size
-        numSteps = numPoints - 1  # arbitrary choice
-        h = 1.0 / numSteps  # compute our step size
-
-        # Compute forward differences from Bezier points and "h"
-        pointX = dx
-        pointY = dy
-
-        firstFDX = ax * (h * h * h) + bx * (h * h) + cx * h
-        firstFDY = ay * (h * h * h) + by * (h * h) + cy * h
-
-        secondFDX = 6 * ax * (h * h * h) + 2 * bx * (h * h)
-        secondFDY = 6 * ay * (h * h * h) + 2 * by * (h * h)
-
-        thirdFDX = 6 * ax * (h * h * h)
-        thirdFDY = 6 * ay * (h * h * h)
-
-        # Compute points at each step
-        result.append((int(pointX), int(pointY)))
-
-        for i in range(numSteps):
-            pointX += firstFDX
-            pointY += firstFDY
-
-            firstFDX += secondFDX
-            firstFDY += secondFDY
-
-            secondFDX += thirdFDX
-            secondFDY += thirdFDY
-
-            result.append((int(pointX), int(pointY)))
-
-        return result
-
     def force_continuity(self, next_point: Vector2) -> Vector2:
         return next_point
 
@@ -454,7 +445,7 @@ class CurveBuilder:
         self.current_points = []
 
     def save_curve(self):
-        new_curves = self.curves_from_points(self.current_curve_type.value, self.current_points)
+        new_curves, _ = self.curves_from_points(self.current_curve_type.value, self.current_points)
         for curve in new_curves:
             self.curves[self.id_provider()] = curve
         self.current_points = []
@@ -502,7 +493,10 @@ class CurveBuilder:
             used_points = points[:-leftover_count] if leftover_count != 0 else points
             curves = self.cubicbeziers_from_points(used_points)
         elif curve_type == CurveType.CATMULLROM:
-            curves = self.catmullrom_from_points(points)
+            leftover_count = len(points) if len(points) < 4 else 0
+            leftover_points = points if leftover_count != 0 else leftover_points
+            used_points = points if leftover_count != 0 else points
+            curves = self.catmullrom_from_points(used_points)
 
         return curves, leftover_points
 
@@ -531,15 +525,19 @@ class CurveBuilder:
             return curves
 
         # A Cubic Bezier requires 4 control points, so operate in blocks of 4 (re-using the last each time)
-        # for ii in range(0, len(points) - 3, 3):
-        #     curves.append(CubicBezier(self.id_provider(), points[ii:ii+4]))
+        for ii in range(0, len(points) - 3, 3):
+            curves.append(CubicBezier(self.id_provider(), points[ii:ii+4]))
 
-        curves.append(CubicBezier(self.id_provider(), points))
+        # curves.append(CubicBezier(self.id_provider(), points))
         return curves
 
     def catmullrom_from_points(self, points):
         curves = []
 
+        if len(points) < 3:
+            return curves
+
+        # A Cubic Bezier requires 4 control points, so operate in blocks of 3 (re-using the last each time)
         curves.append(CatmullRom(self.id_provider(), points))
         return curves
 
@@ -568,30 +566,20 @@ class CurveBuilder:
 
 
 if __name__ == "__main__":
-    # builder = CurveBuilder(lambda: "1")
-    #
-    # a = Vector2(1, 2)
-    # b = Vector2(1, 3)
-    # c = Vector2(1, 4)
-    # d = Vector2(1, 2)
-    # e = Vector2(1, 1)
-    #
-    # builder.set_curve_type(CurveType.CUBICBEZIER)
-    #
-    # builder.add_point(a)
-    # builder.add_point(b)
-    # builder.add_point(c)
-    # builder.add_point(d)
-    # builder.add_point(e)
-    #
-    # not_continuous = Vector2(100.5, 100.5)
-    # continuous = builder.force_continuity(not_continuous)
+    import random as rnd
 
-    builder = CurveBuilder(lambda x: "1")
+    builder = CurveBuilder(lambda: "1")
 
-    not_continuous = Vector2(2, 1)
-    continuous = builder.force_continuity(not_continuous)
-    print(continuous)
+    points = []
 
+    rnd.seed(0)
+    for ii in range(10):
+        points.append(Vector2(rnd.randint(0, 100), rnd.randint(0, 100)))
 
+    curves, leftover = builder.curves_from_points(curve_type=CurveType.CATMULLROM, points=points)
+
+    curve = curves[0]
+    print(curve.control_points)
+    catmullpoints = curves[0].compute_catmull_points(curves[0].control_points)
+    print(catmullpoints)
 
